#include "mpeg2private.h"
#include "mpeg2protos.h"

#include <stdlib.h>

mpeg2_bits_t* mpeg2bits_new_stream(mpeg2_t *file, mpeg2_demuxer_t *demuxer)
{
	mpeg2_bits_t *stream = malloc(sizeof(mpeg2_bits_t));
	stream->bfr = 0;
	stream->value = 0;
	stream->incnt = 0;
	stream->file = file;
	stream->demuxer = demuxer;
	stream->input_ptr = 0;
	return stream;
}

int mpeg2bits_delete_stream(mpeg2_bits_t* stream)
{
	free(stream);
	return 0;
}


inline int mpeg2bits_error(mpeg2_bits_t *stream)
{
	return mpeg2demux_error(stream->demuxer);
}

// Fill a buffer.  Only works if incnt is on an 8 bit boundary
int mpeg2bits_read_buffer(mpeg2_bits_t* stream, unsigned char *buffer, int bytes)
{
	mpeg2_t *file = stream->file;
	int result, i;
	i = 0;
	while(stream->incnt > 0)
	{
		stream->incnt -= 8;
		result = mpeg2demux_read_prev_char(stream->demuxer);
	}

	stream->incnt = 0;
	stream->bfr = 0;
	result = mpeg2demux_read_data(stream->demuxer, buffer, bytes);
	return result;
}

// For mp3 decompression use a pointer in a buffer for getbits.
int mpeg2bits_use_ptr(mpeg2_bits_t* stream, unsigned char *buffer)
{
	stream->incnt = 0;
	stream->bfr = 0;
	stream->input_ptr = buffer;
	return 0;
}

// Go back to using the demuxer for getbits in mp3.
int mpeg2bits_use_demuxer(mpeg2_bits_t* stream)
{
	if(stream->input_ptr)
	{
		stream->input_ptr = 0;
		stream->incnt = 0;
		stream->bfr = 0;
	}

	return 0;
}

// Read bytes backward from the file only until the buffer is full.
int mpeg2bits_fill_stream_backward(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;
	unsigned int result = 0;
	
	if(!stream->incnt) stream->bfr = 0;
	for(; stream->incnt <= 24; stream->incnt += 8)
	{
		result = mpeg2demux_read_prev_char(stream->demuxer);
		stream->bfr |= (result << stream->incnt);
	}
	return result;
}

// Read bytes forward from the file only until the buffer is full.
int mpeg2bits_fill_stream(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;
	unsigned int result = 0;

	if(!stream->incnt) stream->bfr = 0; 
	for(; stream->incnt <= 24; stream->incnt += 8)
	{
		if(stream->input_ptr)
			result = *(stream->input_ptr++);
		else
			result = mpeg2demux_read_char(stream->demuxer);
		stream->bfr |= (result << (24 - stream->incnt)); 
	}
	return result;
}

// Read 4 bytes forward from the file unconditionally.
inline int mpeg2bits_refill(mpeg2_bits_t* stream) 
{
	mpeg2_t *file = stream->file;
	unsigned int result = 0;

	for(stream->bfr = 0, stream->incnt = 0; 
		stream->incnt <= 24 && !mpeg2bits_error(stream); 
		stream->incnt += 8)
	{
		if(stream->input_ptr)
			result = *(stream->input_ptr++);
		else
			result = mpeg2demux_read_char(stream->demuxer);
	  	stream->bfr |= (result << (24 - stream->incnt));
	}
	return mpeg2bits_error(stream);
}

// Read 4 bytes backward from the file unconditionally.
int mpeg2bits_refill_backwards(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;

	for(stream->bfr = 0, stream->incnt = 0; 
		stream->incnt <= 24 && !mpeg2bits_error(stream); 
		stream->incnt += 8)
	  	stream->bfr |= (mpeg2demux_read_prev_char(stream->demuxer) << stream->incnt);
	return mpeg2bits_error(stream);
}

// Move the file position forward by n bits, refilling as needed.
int mpeg2bits_flushbits(mpeg2_bits_t* stream, int n)
{
	if(stream->incnt < n)
	{ 
		n -= stream->incnt; 
		mpeg2bits_refill(stream);
	}
	stream->incnt -= n; 
	stream->bfr <<= n;
	return mpeg2bits_error(stream);
}

// Return n number of bits, advancing the file position.
inline unsigned int mpeg2bits_getbits(mpeg2_bits_t* stream, int n)
{ 
	stream->value = 0;
	if(n > 0)
	{
		if(stream->incnt < n)
		{
			if(stream->incnt)
			{
				n -= stream->incnt;
				stream->value = ((stream->bfr >> (32 - stream->incnt)) << n);
			}
			mpeg2bits_refill(stream);
		}
		stream->value |= (stream->bfr >> (32 - n));
		stream->incnt -= n;
		stream->bfr <<= n;
	}
	return stream->value;
}

// Move the file position backward by n bits, refilling as needed.
int mpeg2bits_flushbits_backwards(mpeg2_bits_t* stream, int n)
{
	int i, total_flushed;
	mpeg2_t *file = stream->file;

	if(n == 0) return 0;

// Put every character in the buffer back
	while(stream->incnt > 0 && !mpeg2bits_error(stream))
	{
		stream->bfr >>= 8;
		stream->incnt -= 8;
		if(stream->input_ptr)
			stream->input_ptr--;
		else
			mpeg2demux_read_prev_char(stream->demuxer);
	}

// Seek back by n bits
	for(i = 0; i < n && !mpeg2bits_error(stream); i += 8)
	{
		if(stream->input_ptr)
			stream->input_ptr--;
		else
			mpeg2demux_read_prev_char(stream->demuxer);
	}

// Read the difference
	stream->incnt = 0;
	if(!mpeg2bits_error(stream)) mpeg2bits_getbits(stream, i - n);

	return mpeg2bits_error(stream);
}

// Fill the buffer only if needed.
// Show the current contents of the buffer up to n bits without advancing the file position.
unsigned int mpeg2bits_showbits(mpeg2_bits_t* stream, int n)
{
	mpeg2bits_fill_stream(stream);
	return (n == 32) ? stream->bfr : (stream->bfr >> (32 - n));
}

int mpeg2bits_byte_align(mpeg2_bits_t *stream)
{
	if(stream->incnt & 7)    
	{ 
		stream->bfr <<= (stream->incnt & 7); 
		stream->incnt -= (stream->incnt & 7); 
	}
	return 0;
}

// Raw seeking methods.

unsigned int mpeg2bits_seek_end(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;

	stream->incnt = 0;
	return mpeg2demux_seek(stream->demuxer, file->total_bytes);
}

unsigned int mpeg2bits_seek_start(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;

	stream->incnt = 0;
	return mpeg2demux_seek(stream->demuxer, 0);
	return 0;
}

unsigned int mpeg2bits_seek_stream(mpeg2_bits_t* stream, long byte_position)
{
	mpeg2_t *file = stream->file;

	stream->incnt = 0;
	return mpeg2demux_seek(stream->demuxer, byte_position);
	return 0;
}

int mpeg2bits_getbitoffset(mpeg2_bits_t *stream)
{
	return stream->incnt & 7;
}

// Higher level routines for video

unsigned int mpeg2bits_prev_startcode(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;

// Byte align & fill bfr
	if(stream->incnt & 7)
	{
		stream->bfr >>= stream->incnt;
		stream->bfr |= mpeg2demux_read_prev_char(stream->demuxer) << 24;
	}
	
// Assume bfr is empty because we just seeked
	mpeg2bits_fill_stream_backward(stream);

// Perform backwards search
	while((stream->bfr >> 8) != MPEG2_PACKET_START_CODE_PREFIX && !mpeg2bits_error(stream))
	{
		stream->bfr >>= 8;
		stream->bfr |= mpeg2demux_read_prev_char(stream->demuxer) << 24;
	}
	return stream->bfr;
}

unsigned int mpeg2bits_startcode(mpeg2_bits_t* stream)
{
	mpeg2_t *file = stream->file;
	unsigned int result = 0;

// Byte align & fill bfr
	mpeg2bits_byte_align(stream);

	result = mpeg2bits_fill_stream(stream);

// Perform search
	while((stream->bfr >> 8) != MPEG2_PACKET_START_CODE_PREFIX && 
		!mpeg2bits_error(stream))  
	{
		stream->bfr <<= 8; 
		result = mpeg2demux_read_char(stream->demuxer);
		stream->bfr |= result;
	}
	return stream->bfr;
}
